#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include "cdl.h"

/*
 * CDLFITS.C -- Routines to load simple FITS files.
 *
 *          cdl_displayFITS  (cdl, fname, frame, fbconfig, zscale)
 *               cdl_isFITS  (fname)
 *             cdl_readFITS  (fname, &pix, &nx, &ny, &bitpix);
 */


#define NCARDS 		36
#define BLOCKSIZE 	2880

/* data types */
enum datatype { T_INT, T_LOG, T_REAL, T_NOVAL };

typedef struct {
	FILE      *fp;              /* file pointer */
	int	  bitpix, size;     /* number of bits per pixel, sizeof(unit) */
	int	  naxis;            /* number of axes */
	long int  axes[2];     	    /* size of each axis */
	long int  ndata;       	    /* number of elements in data */
	long int  cpos;        	    /* current position in data file */
	float     bscale, bzero;    /* scaling parameters */
} FITS;

/* Function prototypes */
#ifdef __STDC__
#include <stddef.h>
#include <stdlib.h>
#endif

#undef max
#define f_max(a,b) ((a) > (b) ? (a) : (b))
#undef min
#define f_min(a,b) ((a) < (b) ? (a) : (b)) 

extern	int	cdl_debug;

static char *cdl_getFITSPixels();
static char *cdl_openFITS(), *cdl_readHeader ();
static char *cdl_rdCard(), *cdl_getData (), *cdl_fixData();



/*  CDL_DISPLAYFITS -- Display a simple FITS format image to the given frame, 
 *  optionally doing a Zscale transform of the image pixels.
 */

int
cdl_displayFITS (cdl, fname, frame, fbconfig, zscale)
CDLPtr  cdl;                            /* package ptr          */
char	*fname;				/* image name		*/
int	frame;				/* display frame	*/
int	fbconfig;			/* frame buffer config	*/
int	zscale;				/* do zscale of image?	*/
{
	char	*errstr;
	int	status, nx, ny, bitpix;
	uchar	*pix;
	float	z1 = 0.0, z2 = 0.0;


	/* See if this is a valid FITS file. */
	if (!cdl_isFITS (fname)) {
	    fprintf (stderr, "%s: not a simple FITS file.\n", fname);
	    return (ERR);
	}

	/* Get the raw FITS image pixels. */
        if (cdl_readFITS (fname, &pix, &nx, &ny, &bitpix))
	    return (ERR);

	if (cdl_debug) {
	    printf("[cdl_displayFITS] '%s' frame=%d zscale=%d\n",
		fname, frame, zscale);
	    printf("[cdl_displayFITS] %dx%d bitpix=%d pixels z1=%g z2=%g\n",
		nx, ny, bitpix, z1, z2);
	}

	status =  cdl_displayPix (cdl, pix, nx, ny, bitpix, frame, fbconfig,
	    zscale);
	free ((char *) pix);
	return (status);
}


/*  CDL_ISFITS -- Test a file to see if it is a simple FITS file.
 */

int 
cdl_isFITS (fname)
char	*fname;				/* input filename */
{
	register FILE *fp;
	int value = 0;
	char keyw[8], val;

	if (fp = fopen (fname, "r")) {
	    fscanf (fp, "%6s = %c", keyw, &val);
	    if (strcmp ("SIMPLE", keyw) == 0 && val == 'T')
		value = 1;
	    fclose (fp);
	}
	return value;
}


/*  CDL_READFITS -- Read the pixels from a simple FITS format image, ret-
 *  urning the pixel array, dimensions, and image type.
 */

int
cdl_readFITS (fname, pix, nx, ny, bitpix)
char	*fname;				/* image name		*/
uchar	**pix;				/* pixel array (output) */
int	*nx, *ny;			/* dimensions (output)	*/
int	*bitpix;			/* pixel size (output)	*/
{
	char	*errstr;
	char	*cdl_getFITSPixels();

	/* See if this is a valid FITS file. */
	if (!cdl_isFITS (fname)) {
	    fprintf (stderr, "%s: not a simple FITS file.\n", fname);
	    return (ERR);
	}

	if (errstr = cdl_getFITSPixels (fname, pix, nx, ny, bitpix)) {
	    fprintf (stderr, "%s\n", errstr);
	    return (ERR);
	}

	if (cdl_debug)
	    printf ("[cdl_readFITS] '%s' nx=%d ny=%d bitpix=%d\n",
		fname, *nx, *ny, *bitpix);

	return (OK);
}


/* ------------------
 * Private Procedures
 * ------------------*/


/*  CDL_GETFITSPIXELS -- Given an FITS filename return a pointer to the pixel
 *  array and the image dimensions.
 */

static char *
cdl_getFITSPixels (fname, pix, nx, ny, bitpix)
char 	*fname;                		/* input filename 	*/
uchar   **pix;                 		/* output pixels 	*/
int 	*nx, *ny;                       /* dimensions 		*/
int	*bitpix;			/* pixel size		*/
{
	FITS 	fs;
	register int	i, nelem;
	char	*error = NULL;


	if (error = cdl_openFITS (&fs, fname, nx, ny, bitpix))
	    return error;

        /* Allocate a buffer to store the image. */
	nelem = (*nx) * (*ny);
        if (fs.bscale != 1.0 && fs.bzero != 0.0)
            *pix = (uchar *) malloc (nelem * f_max(fs.size,4));
        else
            *pix = (uchar *) malloc (nelem * fs.size);
        if (*pix == NULL) 
            return "Insufficient memory for workspace";
        if (error = cdl_getData (&fs, *pix, nelem))
            return error;

        /* If the data is uchar, then read it directly */
        if (fs.bitpix == 8 && (fs.bscale == 1.0 && fs.bzero == 0.0))
            return cdl_getData (&fs, *pix, nelem);

        /* If we've got BSCALE/BZERO values compute the original pixel values
         * and convert the buffer to floating point before processing it.
         * The pix buffer was allocated above with this in mind so it should
         * be large enough that we can fix the pixels in place.
         */
        if (fs.bscale != 1.0 && fs.bzero != 0.0) {
            register float *buf;

            buf = (float *)*pix;

            if (fs.bitpix ==  8) {
                for (i=(nelem-1); i >= 0; i--)
                    buf[i] = (float) (*pix)[i] * fs.bscale + fs.bzero;
            } else if (fs.bitpix ==  16) {
                register short *old = (short *) *pix;
                for (i=(nelem-1); i >= 0; i--)
                    buf[i] = (float) old[i] * fs.bscale + fs.bzero;
            } else if (fs.bitpix ==  32) {
                register int *old = (int *) *pix;
                for (i=(nelem-1); i >= 0; i--)
                    buf[i] = (float) old[i] * fs.bscale + fs.bzero;
            } else if (fs.bitpix ==  -32) {
                register float *old = (float *) *pix;
                for (i=(nelem-1); i >= 0; i--)
                    buf[i] = (float) old[i] * fs.bscale + fs.bzero;
            } else if (fs.bitpix ==  -64) {
                register double *old = (double *) *pix, *buf;
                register float *fpix;

                buf = (double *) malloc (nelem * sizeof(double));
                for (i=(nelem-1); i >= 0; i--)
                    buf[i] = (float) old[i] * fs.bscale + fs.bzero;
                fpix = (float *) *pix;
                for (i=0; i<nelem; i++)
                    fpix[i] = (float) buf[i];
                free ((double *) buf);
            }
        
            fs.size   = 4;
            fs.bitpix = -32;
            *bitpix = -32;
        }

	/* sucess ! */
	fclose(fs.fp);
	return NULL;
}


/*  CDL_OPENFITS -- Open a 2-dimensional FITS file.  Stores the dimensions of
 *  the file in nx and ny, and updates the FITS structure passed in fs.
 *  If successful, returns NULL otherwise returns an error message.
 */

static char *
cdl_openFITS(fs, file, nx, ny, bitpix)
FITS 	*fs;
char	*file;
int	*nx, *ny, *bitpix;
{
	register int	i;
	FILE 	*fp;
	char	*error = NULL;

	if ((fp = fopen(file, "rb")) == NULL) 
	    return "Unable to open FITS file";

	fs->fp = fp;
	fs->bitpix = 0;
	fs->naxis = 0;
	fs->cpos = 0;

	/* Read the header. */
	if (error = cdl_readHeader (fs)) {
	    fclose(fs->fp);
	    return error;
	}

	if (fs->naxis > 2) {
	    fclose(fs->fp);
	    return "Not a 2-D FITS image.";
	}

	/* Get number of pixel. */
	fs->ndata = 1;
	for (i = 0; i < fs->naxis; i++) 
	    fs->ndata = fs->ndata * fs->axes[i];

	*nx = fs->axes[0];
	*ny = fs->axes[1];
	*bitpix = fs->bitpix;

	return NULL;
}


/*  CDL_READHEADER -- Reads the fits header, and updates the FITS structure fs.
 *  Returns NULL on success, or an error message otherwise.
 */

static char *
cdl_readHeader (fs)
FITS *fs;
{
	register int	 i, j, res;
	char	 name[9];
	char	 *block;
	char	 *error = NULL;
	long int val;         /* the value */
	float	 rval;		     /* floating point value */

	if ((block = (char *) malloc(BLOCKSIZE)) == NULL) 
	    return "Insufficient memory for workspace";

	if ((res = fread (block, sizeof(char), BLOCKSIZE, fs->fp)) != BLOCKSIZE)
	    return "Error reading FITS file";
	i = 0;

	/* Read SIMPLE keyword. */
	if (error = cdl_rdCard (block, "SIMPLE", T_LOG, &val, &rval))
	    return error;
	if (val == 0) 
	    return "Not a SIMPLE FITS file";
	i++;

	/* Read BITPIX keyword. */
	if (error = cdl_rdCard(&block[80], "BITPIX", T_INT, &val, &rval))
	    return error;
	if (val != 8 && val != 16 && val != 32 && val != 64 && val != -32 && 
	    val != -64)
	      return "Bad BITPIX value in FITS file";
	fs->bitpix = val;
	j = fs->bitpix;
	if (j < 0) 
	    j = -j;
	fs->size = j / 8;
	i++;

	/* Read NAXIS keyword. */
	if (error = cdl_rdCard(&block[2*80], "NAXIS", T_INT, &val, &rval))
	    return error;
	if (val < 0 || val > 999)
	    return "Bad NAXIS value in FITS file";
	if (val < 2) 
	    return "FITS file is not a two-dimensional image";
	fs->naxis = val;
	i++;

	/* Read NAXISnnn keys.  We allow NAXIS to be > 2 iff the dimensions
 	 * of the extra axes are 1.
 	 */
	for (j = 0; j < fs->naxis; j++) {
	    if (i == NCARDS) {
	    	res = fread(block, sizeof(char), BLOCKSIZE, fs->fp);
	    	if (res != BLOCKSIZE) 
	    	    return "Error reading FITS file";
	    	i = 0;
	    }

	    sprintf (name, "NAXIS%d", j + 1);
	    if (error = cdl_rdCard (&block[i*80], name, T_INT, &val, &rval))
	    	return error;
	    if (val < 0) 
	    	return "Bad NAXISn value in FITS file";
	    if (j < 2) 
	    	fs->axes[j] = val;
	    else if (val != 1) 
	    	return "FITS file is not a two-dimensional image";
	    i++;
	}
	fs->naxis = 2;

	/* Do the remainder. */
	fs->bscale  = 1.0;
	fs->bzero   = 0.0;
	while (1) {

	    /* Try reading a BSCALE or BZERO keyword from this card. */
	    if (block[i*80] == 'B') {
	        error = cdl_rdCard(&block[i*80], "BSCALE", T_REAL, &val, &rval);
	        if (error == NULL)
	            fs->bscale = rval;
	        error = cdl_rdCard (&block[i*80], "BZERO", T_REAL, &val, &rval);
	        if (error == NULL)
	            fs->bzero = rval;
	    }

	    if (i == NCARDS) {
	    	res = fread (block, sizeof(char), BLOCKSIZE, fs->fp);
	    	if (res != BLOCKSIZE) 
	    	    return "Unexpected EOF in FITS file";
	    	i = 0;
	    }
	    if (strncmp (&block[i*80], "END     ", 8) == 0) 
	    	break;
	    i++;
	}

	free (block);
	return NULL;
}


/*  CDL_RDCARD -- Read a header record, from the 80 byte buffer card.
 *  The keyword name must match 'name'; and parse its value according to
 *  dtype. This can have the following values:
 *      dtype = T_LOG      # value is logical, either 'T' or 'F' in column 30
 *      dtype = T_INT      # value is an int, right justified in columns 11-30
 *      dtype = T_REAL     # value is a real
 *
 *  The value is stored in kvalue.  Returns NULL on success, or an error
 *  message otherwise.
 */

static char *
cdl_rdCard (card, name, dtype, kvalue, rvalue)
char	*card;				/* FITS keyword card 	*/
char	*name;				/* keyword name		*/
enum datatype dtype;   			/* type of value 	*/
long 	*kvalue;			/* integer value 	*/
float 	*rvalue;			/* real value 		*/
{
	register int	i, ptr;
	char	namestr[9];
	char	error[45];

	/* Get the keyword from the card.  */
	bcopy (card, namestr, 8);
        for (i=7; i >= 0 && namestr[i] == ' '; i--)
	    ;
	namestr[i+1] = '\0';
	if (strcmp(namestr, name) != 0)
	    return "Keyword not found in FITS file.";

	/* Get start of value. */
	ptr = 10;
	while (ptr < 80 && card[ptr] == ' ') 
	    ptr++;
	if (ptr == 80) 
	    return "FITS file has missing keyword value"; /* no value */

	if (dtype == T_LOG) {
	    if (ptr != 29 || (card[29] != 'T' && card[29] != 'F'))
	    	return "Keyword has bad logical value in FITS file";
	    *kvalue = (card[29] == 'T');

	} else {
	    /* An integer or real value. */
	    long int	ival;
	    float	fval;
	    char	num[21];

	    if (ptr > 29) 
	    	return "Keyword has bad integer value in FITS file";
	    memcpy (num, &card[ptr], 30 - ptr);
	    num[30-ptr] = '\0';
	    if (dtype == T_INT) {
	       i = sscanf (num, "%ld", &ival);
	       if (i != 1) 
	    	   return "Keyword has bad integer value in FITS file";
	       *kvalue = ival;
	       *rvalue = 0.0;
	    } else if (dtype == T_REAL) {
	       i = sscanf (num, "%g", &fval);
	       if (i != 1) 
	    	   return "Keyword has bad real value in FITS file";
	       *kvalue = 0;
	       *rvalue = fval;
	    }
	}

	return NULL;
}


/*  CDL_GETDATA -- Reads nelem values into the buffer.  Copes with the fact
 *  that the last 2880 byte record of the FITS file may be truncated, and
 *  should be padded out with zeros.  Returns NULL for success or an error
 *  message.
 */

static char *
cdl_getData (fs, buffer, nelem)
FITS 	*fs;
uchar 	*buffer;
int	nelem;
{
	int	res;

	if (nelem == 0) 
	    return NULL;

	if ((res = fread (buffer, fs->size, nelem, fs->fp)) != nelem) {
	    /* Nblock is the number of elements in a record. size is
	     * always a factor of BLOCKSIZE
	     */
	    int	loffs, nblock = BLOCKSIZE / fs->size;

	    if (!feof(fs->fp)) 
	    	return "I/O error reading FITS file";

	    /* The last record might be short; check this.
 	     * loffs is the offset of the start of the last record from
	     * the current position.
 	     */
	    loffs = ((fs->ndata + nblock - 1) / nblock - 1) * nblock - fs->cpos;

	    /* If we didn't read to the end of the penultimate record */
	    if (res < loffs) 
	    	return "Unexpected eof reading FITS file";

	    /* Pad with zeros */
	    memset ((char *)buffer+res*fs->size, '\0', (nelem-res)*fs->size);
	}

	fs->cpos += res;
	return cdl_fixData(fs, buffer, nelem);
}


/*  CDL_FIXDATA -- convert the raw data, as stored in the FITS file, to the
 *  format appropiate for the data representation of the host computer.
 *  Assumes that
 *       short int = 2 byte integer
 *       int       = 4 byte integer
 */

static char *
cdl_fixData (fs, buffer, nelem)
FITS 	*fs;
void 	*buffer;
int	nelem;
{
	register int	i, n = nelem;
	register uchar	*ptr = buffer;

	/* Conversions. Although the data may be signed, reverse using unsigned 
         * variables.  Convert from big-endian two-byte signed integer to
	 * native form.
	 */
	if (fs->bitpix == 16)
	    for (i = 0; i < n; i++, ptr += 2)
	    	*(unsigned short int *)ptr = (((int)*ptr) << 8) | (int)(ptr[1]);

	/* Convert from big-endian four-byte signed integer to native form */
	else if (fs->bitpix == 32)
	    for (i = 0; i < n; i++, ptr += 4)
	    	*(unsigned int *)ptr = (((unsigned int)*ptr) << 24) | 
	    	    ((unsigned int)ptr[1] << 16) | 
	    	    ((unsigned int)ptr[2] << 8)  | 
	    	    (unsigned int)ptr[3];

	/* Convert from IEE 754 single precision to native form */
	else if (fs->bitpix == -32) {
	    register int	j, k, expo;
	    static float	*exps = NULL;

	    if (exps == NULL) {
	    	exps = (float *)calloc(256, sizeof(float));
	    	if (exps == NULL) 
	    	    return "Insufficient memory for workspace";
	    	exps[150] = 1.;
	    	for (i = 151; i < 256; i++) 
	    	    exps[i] = 2. * exps[i-1];
	    	for (i = 149; i >= 0; i--) 
	    	    exps[i] = 0.5 * exps[i+1];
	    }


	    for (i = 0; i < n; i++, ptr += 4) {
	    	k = (int)*ptr;
	    	j = ((int)ptr[1] << 16) | ((int)ptr[2] << 8) | (int)ptr[3];
	    	expo = ((k & 127) << 1) | (j >> 23);
	    	if ((expo | j) == 0) 
	    	    *(float *)ptr = 0.;
	    	else
	    	    *(float *)ptr = exps[expo] * (float)(j | 0x800000);
	    	if (k & 128) 
	    	    *(float *)ptr = -*(float *)ptr;
	    }

	/* Convert from IEE 754 double precision to native form */
	} else if (fs->bitpix == -64) {
	    register int	expo, k, l;
	    register unsigned int	j;
	    static double	*exps = NULL;

	    if (exps == NULL) {
	    	exps = (double *)calloc(2048, sizeof(double));
	    	if (exps == NULL) 
	    	    return "Insufficient memory for workspace";
	    	exps[1075] = 1.;
	    	for (i = 1076; i < 2048; i++) 
	    	    exps[i] = 2. * exps[i-1];
	    	for (i = 1074; i >= 0; i--) 
	    	    exps[i] = 0.5 * exps[i+1];
	    }

	    for (i = 0; i < n; i++, ptr += 8) {
	    	k = (int)*ptr;
	    	j = ((unsigned int)ptr[1] << 24) | ((unsigned int)ptr[2] << 16)|
	    	    ((unsigned int)ptr[3] << 8) | (unsigned int)ptr[4];
	    	l = ((int)ptr[5] << 16) | ((int)ptr[6] << 8) | (int)ptr[7];
	    	expo = ((k & 127) << 4) | (j >> 28);
	    	if ((expo | j | l) == 0) 
	    	    *(double *)ptr = 0.;
	    	else
	    	    *(double *)ptr = exps[expo] * (16777216. * 
	    	        (double)((j & 0x0FFFFFFF) | 0x10000000) + (double)l);
	    	if (k & 128) 
	    	    *(double *)ptr = -*(double *)ptr;
	    }
	}

	return NULL;
}
