#include <stdio.h>
#ifdef ULTRIX
#include <sys/types.h>
#endif
#include <unistd.h>
#include "cdl.h"

/*
 * CDLIRAF.C -- Routines to read IRAF OIF images.
 *
 *          cdl_displayIRAF  (cdl, fname, band, frame, fbconfig, zscale)
 *               cdl_isIRAF  (fname)
 *             cdl_readIRAF  (fname, band, &pix, &nx, &ny, &bitpix);
 */


/* Image header parmeters. */
#define SZ_IMPIXFILE 	79
#define LEN_IMHDR 	513
#define SZ_IMHDR 	2048

/* Offsets into header (in sizeof(int) units) for various parameters */
#define	IM_PIXTYPE	4		/* datatype of the pixels */
#define	IM_NDIM		5		/* number of dimensions */
#define	IM_LEN		6		/* image dimensions */
#define	IM_PHYSLEN	13		/* physical length (as stored) */
#define	IM_PIXOFF	22		/* offset of the pixels */
#define	IM_PIXFILE	103		/* name of pixel storage file */

/* IRAF dataype codes */
#define	TY_CHAR		2
#define	TY_SHORT	3
#define	TY_INT		4
#define	TY_LONG		5
#define	TY_REAL		6
#define	TY_DOUBLE	7
#define	TY_USHORT	11
#define	TY_UCHAR	12

#ifdef __STDC__
#include <stddef.h>
#include <stdlib.h>
#endif

#undef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#undef min
#define min(a,b) ((a) < (b) ? (a) : (b))

extern	int	cdl_debug;

static char *cdl_getIRAFPixels();
static void  cdl_getPixfile(), cdl_strpak();



/*  CDL_DISPLAYIRAF -- Display an IRAF OIF format image to the given frame,
 *  optionally doing a Zscale transform of the image pixels.
 */

int
cdl_displayIRAF (cdl, fname, band, frame, fbconfig, zscale)
CDLPtr  cdl;                            /* package ptr          */
char	*fname;				/* image name		*/
int	band;				/* image band if 3-d	*/
int	frame;				/* display frame	*/
int	fbconfig;			/* frame buffer config	*/
int	zscale;				/* do zscale of image?	*/
{
	char	*errstr;
	int	status, nx, ny, bitpix;
	uchar	*pix = NULL;
	float	z1 = 0.0, z2 = 0.0;


	/* See if this is a valid IRAF file. */
	if (!cdl_isIRAF (fname)) {
	    fprintf (stderr, "%s: not a valid IRAF file.\n", fname);
	    return (ERR);
	}

	/* Get the raw IRAF image pixels. */
        if (cdl_readIRAF (fname, band, &pix, &nx, &ny, &bitpix))
	    return (ERR);

	if (cdl_debug) {
	    printf("[cdl_displayIRAF] '%s[%d]' frame=%d zscale=%d\n",
		fname, band, frame, zscale);
	    printf("[cdl_displayIRAF] %dx%d bitpix=%d pixels z1=%g z2=%g\n",
		nx, ny, bitpix, z1, z2);
	}

	status =  cdl_displayPix (cdl, pix, nx, ny, bitpix, frame, fbconfig,
	    zscale);
	free ((char *) pix);
	return (status);
}


/* CDL_ISIRAF -- Test a file to see if it is a IRAF file.
 */

int 
cdl_isIRAF (fname)
char	*fname;				/* input filename */
{
	register FILE *fp;
	int value = 0;
	char magic[24];

        if (fp = fopen (fname, "r")) {
            fread (magic, sizeof (char), 12, fp);
	    cdl_strpak (magic, magic, 5);
	    if (strncmp(magic, "imhdr", 5) == 0)
	        value = 1;
	    fclose (fp);
	}

	return value;
}


/*  CDL_READIRAF -- Read the pixels from an IRAF OIF format image, returning
 *  the pixel array, dimensions, and image type.
 */

int
cdl_readIRAF (fname, band, pix, nx, ny, bitpix)
char	*fname;				/* image name		*/
int	band;				/* image band if 3-d	*/
uchar	**pix;				/* pixel array (output) */
int	*nx, *ny;			/* dimensions (output)	*/
int	*bitpix;			/* pixel size (output)	*/
{
	char	*errstr;
	char	*cdl_getIRAFPixels();


	/* See if this is a valid IRAF file. */
	if (!cdl_isIRAF (fname)) {
	    fprintf (stderr, "%s: not a valid IRAF file.\n", fname);
	    return (ERR);
	}

	if (errstr = cdl_getIRAFPixels (fname, band, pix, nx, ny, bitpix)) {
	    fprintf (stderr, "%s\n", errstr);
	    return (ERR);
	}

	if (cdl_debug)
	    printf ("[cdl_readIRAF] '%s[%d]' nx=%d ny=%d bitpix=%d\n",
		fname, band, *nx, *ny, *bitpix);

	return (OK);
}



/* ------------------
 * Private routines.
 * ------------------*/

/*  CDL_GETIRAFPIXELS -- Given an IRAF filename return a pointer to the pixel
 *  array and the image dimensions.
 */

static char *
cdl_getIRAFPixels (fname, band, image, nx, ny, bitpix)
char    *fname;                         /* input filename       */
int	band;				/* image band		*/
uchar   **image;                        /* output pixels        */
int     *nx, *ny;                       /* dimensions           */
int	*bitpix;			/* pixel size		*/
{
	register int i, npix, px, py, nbands, pix_size;
	FILE 	*hdr, *pf;
	char  	*line;

	int header[SZ_IMHDR], ptype, offset;
	char path[SZ_IMPIXFILE], pixfile[SZ_IMPIXFILE];

        if ((hdr = fopen (fname, "r")) == NULL)
	    return "cannot open image";

	/* Read in the image header. */
        fread (header, sizeof (char), SZ_IMHDR, hdr);

	/* Get the interesting stuff. */
	px = header[IM_PHYSLEN];
	py = header[IM_PHYSLEN+1];
	*nx = header[IM_LEN];
	*ny = header[IM_LEN+1];
	ptype = header[IM_PIXTYPE];
	offset = (header[IM_PIXOFF] - 1) * sizeof(short);
	nbands = header[IM_LEN+2];
	if (band > nbands)
	    return "Invalid band request.";

	/* Find the pixfile and see if it exists. */
	cdl_strpak ((char *)&header[IM_PIXFILE], pixfile, SZ_IMPIXFILE);
	cdl_getPixfile (fname, pixfile);
	if (access (pixfile, R_OK) != 0) {
	    fclose (hdr);	
	    return "Cannot access pixel file";
	}
	fclose (hdr); 			/* we're done with the header */


	/* Open the pixel file and seek to the beginning of the data. */
        if ((pf = fopen (pixfile, "r")) == NULL)
	    return "Cannot open pixel file.";

	npix = (*nx) * (*ny);
	switch (ptype) {
	case TY_USHORT:
	    pix_size = sizeof (unsigned short);
	    *bitpix = 16;
	    break;
	case TY_SHORT:
	    pix_size = sizeof (short);
	    *bitpix = 16;
	    break;
	case TY_INT:
	    pix_size = sizeof (int);
	    *bitpix = 32;
	    break;
	case TY_LONG:
	    pix_size = sizeof (long);
	    *bitpix = 32;
	    break;
	case TY_REAL:
	    pix_size = sizeof (float);
	    *bitpix = -32;
	    break;
	case TY_DOUBLE:
	    pix_size = sizeof (double);
	    *bitpix = -64;
	    break;
	default:
	    fclose (pf);
	    return "Invalid pixel type in image";
	}

	/* Now suck up the pixels. */
	*image = (uchar *) malloc (npix * pix_size);
	offset += (band - 1) * (px * py) * pix_size;
	lseek (fileno(pf), (off_t)offset, SEEK_SET);

	if (*nx == px)
            fread (*image, pix_size, npix, pf);
	else {
	    line = (char *) malloc (px * pix_size);
	    for (i=0; i < (*ny); i++) {
                fread (line, pix_size, px, pf);
		bcopy (line, &((*image)[i*(*nx)*pix_size]), (*nx)*pix_size);
	    }
	    free ((char *)line);
	}

	fclose (pf);
	return NULL;
}


/*  CDL_GETPIXFILE -- Get the pixelfile pathname.
 */

static void
cdl_getPixfile (fname, pfile)
char	*fname;
char	*pfile;
{
	char	temp[SZ_IMPIXFILE], *ip;
	int	len;
	char  	*index();

        if (strncmp (pfile, "HDR$", 4) == 0) {
            /* Handle the special case of a HDR$ pixfile path, prepend the
             * current working directory name to the pixfile.
             */
	    ip = pfile + 4;
            (void) strncpy (temp, ip, SZ_IMPIXFILE);
            (void) strncpy (pfile, fname, SZ_IMPIXFILE);

            /* Find the end of the pathname. */
            len = strlen (pfile);
            while ( (len > 0) && (pfile[len-1] != '/') )
              len--;

            /* Add the image name. */
            pfile[len] = '\0';
            (void) strncat (pfile, temp, SZ_IMPIXFILE);

        } else if (index (pfile, '!') != NULL) {
            /* Strip out the leading node! prefix from the pixfile path.  */
            for (ip = pfile; *ip != '!' ; )
                ip++;
            (void) strcpy (pfile, ++ip);
        }
}


/*  CDL_STRPAK -- Convert ASCII string from SPP char per short to C char
 *  per byte.
 */

static void 
cdl_strpak (in, out, len)
char	*in, *out;
int	len;
{
	int	i, j;

	/* Adaptive byte selection (whichever byte order) chars alternate 
	 * with \0.
	 */
	j = (in[0] == '\0' ? 1 : 0);

	for (i = 0; i < len; i++, j += 2)
	    out[i] = in[j];
	out[i] = '\0';
}
