#include <stdio.h>
#include <math.h>
#include "cdl.h"
#include "cdlfont.h"

/*
 *  CDL Marker Procedures -- Procedures for doing overlay graphics.
 *
 *             cdl_mapFrame  (cdl, frame)
 *            cdl_markPoint  (cdl, x, y, number, size, type, color)
 *             cdl_markLine  (cdl, xs, ys, xe, ye, color)
 *              cdl_markBox  (cdl, lx, ly, ux, uy, fill, color)
 *         cdl_markPolyline  (cdl, xpts, ypts, npts, color)
 *          cdl_markPolygon  (cdl, xpts, ypts, npts, fill, color)
 *           cdl_markCircle  (cdl, x, y, radius, fill, color)
 *       cdl_markCircAnnuli  (cdl, x, y, radius, nannuli, sep, color)
 *          cdl_markEllipse  (cdl, x, y, xrad, yrad, ang, fill, color)
 *      cdl_markEllipAnnuli  (cdl, x, y, xrad, yrad, ang, nannuli, sep, color)
 *             cdl_markText  (cdl, x, y, str, size, angle, color)
 *
 *           cdl_deleteMark  (cdl, x, y)
 *         cdl_clearOverlay  (cdl)
 *        cdl_redrawOverlay  (cdl)
 */

/* Function prototypes */
#ifdef __STDC__
#include <stddef.h>
#include <stdlib.h>
#endif

/* Display list declarations.  We keep a separate list for each frame that
 * is freed whenever a new image is displayed.  The list is maintained as
 * a doubly-linked list of Marker structs.
 */
MarkerPtr  DLHead[MAX_FRAMES];          /* diplay list head     */
MarkerPtr  DLTail[MAX_FRAMES];          /* diplay list tail     */


/* absolute value of a */
#define ABS(a)          (((a)<0) ? -(a) : (a))

/* take binary sign of a, either -1, or 1 if >= 0 */
#define SGN(a)          (((a)<0) ? -1 : 1)

#define	N_CIRCLE_PTS	48	/* no. of points defining unit circle  	*/
#define	N_ELLIPSE_PTS	64	/* no. of points defining unit ellipse 	*/
#define FONT_SCALE      0.33	/* default text font scale		*/


extern int	cdl_debug;

int		cdl_freeDisplayList();

static int	cdl_getPixRegion();
static void	cdl_insertMarker(), cdl_removeMarker(); 
static void	cdl_doLineInSubRas();
static void	cdl_doPointMark(), cdl_doLineMark(), cdl_doBoxMark();
static void	cdl_doPolygonMark(), cdl_doCircleMark(), cdl_doCircAnnuliMark();
static void	cdl_doEllipseMark(), cdl_doEllipAnnuliMark();
static void	cdl_getCircleCoords(), cdl_getEllipseCoords();
static void	cdl_printDisplayList(), cdl_initMarkPos();
static void	cdl_fillArea(), cdl_minmax();

static int	bitupk();

static MarkerPtr cdl_initPointMarker(), cdl_initLineMarker();
static MarkerPtr cdl_initBoxMarker(), cdl_initPolylineMarker();
static MarkerPtr cdl_initPolygonMarker(), cdl_initCircleMarker();
static MarkerPtr cdl_initCircAnnMarker(), cdl_initEllipseMarker();
static MarkerPtr cdl_initEllAnnMarker(), cdl_findNearest();
static MarkerPtr cdl_initTextMarker();



/*  CDL_MAPFRAME -- Map the current frame buffer as an image for overlay.
 *  This routine is a bit of a hack since we're trying to make some guesses
 *  about what the server has displayed, there's no way to actually query it.
 *  We assume the fbconfig has already been reset if it's not the default,
 *  and we assume the WCS defined is for a single image centered in the frame.
 */

int
cdl_mapFrame (cdl, frame)
CDLPtr	cdl;				/* package ptr		*/
int	frame;				/* fb frame to map	*/
{
	int	xo, yo;
	float   a, b, c, d, tx, ty, z1, z2;
	int     zt;
	char	*name = NULL, *title = NULL;

        name  = (char *) malloc (SZ_NAME);
        title = (char *) malloc (SZ_NAME);

	/* Select the requested frame. */
	cdl_setFrame (cdl, frame);

	/* Get/Set the current WCS. */
	(void) cdl_getWCS (cdl, name, title, &a, &b, &c, &d, &tx, &ty,
	    &z1, &z2, &zt);

	/* If no WCS is defined for the frame make up a default. */
	if ((a+b+c+d+tx+ty+z1+z2+zt) == 0.0) {
            a  =  1.0;
            b  =  0.0;
            c  =  0.0;
            d  = -1.0;
            tx =  1.0;
            ty = (float) cdl->fbheight;
            z1 =  0.0;
            z2 = 255.0;
            zt = CDL_LINEAR;
	}
	(void) cdl_setWCS (cdl, name, title, a, b, c, d, tx, ty, z1, z2, zt);

	/* Reset the image corners, assume image is centered in the frame.  */
	xo = - cdl->tx + 1;
	yo = cdl->fbheight - cdl->ty;
	cdl->imd->xs = xo;
	cdl->imd->ys = yo;
	cdl->imd->xe = cdl->fbwidth - xo + 1;
	cdl->imd->ye = cdl->fbheight - yo + 1;
}


/*  CDL_MARKPOINT --  Draw a mark on the display at the given point.  Marks
 *  be specified as either a '+', 'x', or '*'.  If the number argument is
 *  positive that value will be drawn next to the mark as a label string.
 */

int
cdl_markPoint (cdl, x, y, number, size, type, color)
CDLPtr	cdl;				/* package ptr		*/
int	x, y;				/* marker position	*/
int	number;				/* if > 0, label value	*/
int	size;				/* marker size (pixels) */
int	type;				/* type to draw		*/
int	color;				/* marker color		*/
{
	register int lx, ly, sz;
	uchar	*pix = NULL;
	MarkerPtr mk = cdl_initPointMarker (x, y, number, size, type, color);

	if (cdl_debug)
	    printf ("[cdl_markPoint] (%dx%d) num=%d size=%d type=%d color=%d\n",
		x, y, number, size, type, color);

	sz = (size % 2 ? size : size + 1);
	sz = max (sz, 3);
	lx = x - (sz / 2);
	ly = y - (sz / 2);
	cdl_initMarkPos (mk, sz, sz, lx, ly);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (sz * sz);
	if (cdl_getPixRegion (cdl, lx, ly, sz, sz, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (sz * sz);
	mk->markpix = (uchar *) malloc (sz * sz);
	bcopy (pix, mk->refpix, sz * sz);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster. */
	cdl_doPointMark (color, type, pix, sz);
	mk->markpix = pix;

	/* Number the point if requested. */
	if (number > 0) {
	    char   lab[64];
	    int	   cx, cy, llen;

	    sprintf (lab, "%d", number);
	    llen = strlen(lab) * (CHARACTER_WIDTH * FONT_SCALE) + strlen(lab);

	    /* Position the label so it isn't clipped. */
	    cx = x + (sz / 2) - 1;
	    cy = y + (sz / 2) - 1;
	    if ((cx + llen) > cdl->im_nx)   cx = x - llen + 1;
	    if ((cy + llen) > cdl->im_ny)   cy = y - llen + 1;
	    cdl_markText (cdl, cx, cy, lab, 1., 0.0, color);
	}

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, sz, sz, pix))
	    return (ERR);

	return (OK);
}


/*  CDL_MARKLINE --  Draw a line of given color between the two points
 *  specified.
 */

int
cdl_markLine (cdl, xs, ys, xe, ye, color)
CDLPtr	cdl;				/* package ptr		*/
int	xs, ys;				/* line start points	*/
int	xe, ye;				/* line end points	*/
int	color;				/* marker color		*/
{
	register int i, j, tmp, nx, ny, lx, ly;
	uchar	*pix = NULL;
	MarkerPtr mk = cdl_initLineMarker (xs, ys, xe, ye, color);

	if (cdl_debug)
	    printf ("[cdl_markLine] (%d,%d) -> (%d,%d) color=%d\n",
		xs, ys, xe, ye, color);

	nx = ABS(xe-xs) + 1;   lx = min(xs,xe);
	ny = ABS(ye-ys) + 1;   ly = min(ys,ye);;
	cdl_initMarkPos (mk, nx, ny, lx, ly);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (nx * ny);
	if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (nx * ny);
	mk->markpix = (uchar *) malloc (nx * ny);
	bcopy (pix, mk->refpix, nx * ny);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster. */
	cdl_doLineMark (color, pix, xs, ys, xe, ye);
	mk->markpix = pix;

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	return (OK);
}


/*  CDL_MARKBOX --  Draw a rectangular box given two corner endpoints.  The
 *  box may be optionally filled with the specified color.
 */

int
cdl_markBox (cdl, lx, ly, ux, uy, fill, color)
CDLPtr	cdl;				/* package ptr		*/
int	lx, ly;				/* LL corner points	*/
int	ux, uy;				/* UR corner points	*/
int	fill;				/* fill rectangle?	*/
int	color;				/* marker color		*/
{
	register int i, j, tmp, nx, ny;
	uchar	*pix = NULL;
	MarkerPtr mk = cdl_initBoxMarker (lx, ly, ux, uy, fill, color);

	if (cdl_debug)
	    printf ("[cdl_markBox] (%d,%d) -> (%d,%d) fill=%d color=%d\n",
		lx, ly, ux, uy, fill, color);

	/* Take care of a corner specified the wrong way. */
	if (lx > ux) { tmp = ux;   ux = lx;  lx = tmp; }
	if (ly > uy) { tmp = uy;   uy = ly;  ly = tmp; }
	nx = ux - lx + 1;
	ny = uy - ly + 1;
	cdl_initMarkPos (mk, nx, ny, lx, ly);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (nx * ny);
	if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (nx * ny);
	mk->markpix = (uchar *) malloc (nx * ny);
	bcopy (pix, mk->refpix, nx * ny);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster. */
	cdl_doBoxMark (color, fill, pix, nx, ny);
	mk->markpix = pix;

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	return (OK);
}


/*  CDL_MARKPOLYGON -- Draw a polygon on the display, optionally filling with
 *  the specified color.
 */

int
cdl_markPolygon (cdl, xarray, yarray, npts, fill, color)
CDLPtr	cdl;				/* package ptr		*/
int	xarray[], yarray[];		/* vertex points	*/
int	npts;				/* number of corners	*/
int	fill;				/* fill polygon?	*/
int	color;				/* marker color		*/
{
	register int nx, ny, lx, ly;
	int 	xmin, xmax, ymin, ymax;
	uchar 	*pix;
	MarkerPtr mk = cdl_initPolygonMarker(xarray, yarray, npts, fill, color);

	/* Find the boundaries of the polygon. */
	cdl_minmax (xarray, npts, &xmin, &xmax);
	cdl_minmax (yarray, npts, &ymin, &ymax);
	nx = xmax - xmin + 1;   lx = xmin;
	ny = ymax - ymin + 1;   ly = ymin;
	cdl_initMarkPos (mk, nx, ny, lx, ly);

	if (cdl_debug)
	    printf ("[cdl_markPolygon] npts=%d fill=%d color=%d nx=%d ny=%d\n",
		npts, fill, color, nx, ny);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (nx * ny);
	if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (nx * ny);
	mk->markpix = (uchar *) malloc (nx * ny);
	bcopy (pix, mk->refpix, nx * ny);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster. */
	cdl_doPolygonMark (color, fill, xarray, yarray, npts,
	    pix, lx, ly, nx, ny, True);
	mk->markpix = pix;

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	return (OK);
}


/*  CDL_MARKPOLYLINE -- Draw a polyline on the display, optionally filling with
 *  the specified color.
 */

int
cdl_markPolyline (cdl, xarray, yarray, npts, color)
CDLPtr	cdl;				/* package ptr		*/
int	*xarray, *yarray;		/* vertex points	*/
int	npts;				/* number of points	*/
int	color;				/* marker color		*/
{
	register int nx, ny, lx, ly;
	int 	xmin, xmax, ymin, ymax;
	uchar 	*pix;
	MarkerPtr mk = cdl_initPolylineMarker (xarray, yarray, npts, color);

	/* Find the boundaries of the polyline. */
	cdl_minmax (xarray, npts, &xmin, &xmax);
	cdl_minmax (yarray, npts, &ymin, &ymax);
	nx = xmax - xmin + 1;   lx = xmin;
	ny = ymax - ymin + 1;   ly = ymin;
	cdl_initMarkPos (mk, nx, ny, lx, ly);

	if (cdl_debug)
	    printf ("[cdl_markPolyline] npts=%d color=%d nx=%d ny=%d\n",
		npts, color, nx, ny);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (nx * ny);
	if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (nx * ny);
	mk->markpix = (uchar *) malloc (nx * ny);
	bcopy (pix, mk->refpix, nx * ny);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster.  */
	cdl_doPolygonMark (color, False, xarray, yarray, npts,
	    pix, lx, ly, nx, ny, False);
	mk->markpix = pix;

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	return (OK);
}


/*  CDL_MARKCIRCLE -- Draw of circle on the display, optionally filling with
 *  the specified color.
 */

int
cdl_markCircle (cdl, x, y, radius, fill, color)
CDLPtr	cdl;				/* package ptr		*/
int	x, y;				/* center position	*/
int	radius;				/* radius of circle	*/
int	fill;				/* fill circle?		*/
int	color;				/* marker color		*/
{
	register int nx, ny, lx, ly;
	int 	xmin, xmax, ymin, ymax;
	uchar 	*pix;
	MarkerPtr mk = cdl_initCircleMarker (x, y, radius, fill, color);

	nx = ny = 2 * radius + 3;
	lx = x - radius - 1;
	ly = y - radius - 1;
	cdl_initMarkPos (mk, nx, ny, lx, ly);

	if (cdl_debug)
	    printf ("[cdl_markCircle] (%d,%d) radius=%d fill=%d color=%d\n",
		x, y, radius, fill, color);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (nx * ny);
	if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (nx * ny);
	mk->markpix = (uchar *) malloc (nx * ny);
	bcopy (pix, mk->refpix, nx * ny);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster. */
	cdl_doCircleMark (x, y, radius, color, fill, pix, lx, ly, nx, ny);
	mk->markpix = pix;

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	return (OK);
}


/*  CDL_MARKCIRCANNULI -- Draw circular annuli on the display.
 */

int
cdl_markCircAnnuli (cdl, x, y, radius, nannuli, sep, color)
CDLPtr	cdl;				/* package ptr		*/
int	x, y;				/* center position	*/
int	radius;				/* radius of 1st annulus*/
int	nannuli;			/* no. of annuli	*/
int	sep;				/* annuli sep (pixels)	*/
int	color;				/* marker color		*/
{
	register int nx, ny, lx, ly;
	int 	xmin, xmax, ymin, ymax;
	uchar 	*pix;
	MarkerPtr mk = cdl_initCircAnnMarker(x, y, radius, nannuli, sep, color);

	nx = ny = 2 * (radius + nannuli * sep) + 3;
	lx = x - (nx / 2);
	ly = y - (ny / 2);
	cdl_initMarkPos (mk, nx, ny, lx, ly);

	if (cdl_debug)
	    printf("[cdl_markAnnuli] (%d,%d) radius=%d N=%d sep=%d color=%d\n",
		x, y, radius, nannuli, sep, color);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (nx * ny);
	if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (nx * ny);
	mk->markpix = (uchar *) malloc (nx * ny);
	bcopy (pix, mk->refpix, nx * ny);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster. */
	cdl_doCircAnnuliMark (x, y, radius, nannuli, sep, color,
	    pix, lx, ly, nx, ny);
	mk->markpix = pix;

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	return (OK);
}


/*  CDL_MARKELLIPSE -- Draw an ellipse with a given center, eccentricity
 *  and position on the display, optionally filling with the specified color.
 */

int
cdl_markEllipse (cdl, x, y, xrad, yrad, rotang, fill, color)
CDLPtr	cdl;				/* package ptr		*/
int	x, y;				/* center position	*/
int	xrad, yrad;			/* x and y radii	*/
float	rotang;				/* position angle (deg) */
int	fill;				/* fill ellipse?	*/
int	color;				/* marker color		*/
{
	register int nx, ny, lx, ly;
	int 	xmin, xmax, ymin, ymax;
	uchar 	*pix;
	MarkerPtr mk = cdl_initEllipseMarker (x, y, xrad, yrad, rotang,
	    fill, color);

	nx = 2 * max(xrad,yrad) + 3;
	ny = 2 * max(xrad,yrad) + 3;
	lx = x - (nx / 2);
	ly = y - (ny / 2);
	cdl_initMarkPos (mk, nx, ny, lx, ly);

	if (cdl_debug)
	    printf("[cdl_markEllipse] (%d,%d) R=(%d,%d) pa=%g fill=%d col=%d\n",
		x, y, xrad, yrad, rotang, fill, color);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (nx * ny);
	if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (nx * ny);
	mk->markpix = (uchar *) malloc (nx * ny);
	bcopy (pix, mk->refpix, nx * ny);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster. */
	cdl_doEllipseMark (color, fill, x, y, xrad, yrad, rotang,
	    pix, lx, ly, nx, ny);
	mk->markpix = pix;

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	return (OK);
}


/*  CDL_MARKELLIPANNULI -- Draw elliptical annuli on the display.
 */

int
cdl_markEllipAnnuli (cdl, x, y, xrad, yrad, ang, nannuli, sep, color)
CDLPtr	cdl;				/* package ptr		*/
int	x, y;				/* center position	*/
int	xrad, yrad;			/* radius of 1st annulus*/
float	ang;				/* rotation angle	*/
int	nannuli;			/* no. of annuli	*/
int	sep;				/* annuli sep (pixels)	*/
int	color;				/* marker color		*/
{
	register int nx, ny, lx, ly;
	uchar 	*pix;
	MarkerPtr mk = cdl_initEllAnnMarker (x, y, xrad, yrad, ang, nannuli,
	    sep, color);

	nx = 2 * max(xrad + nannuli * sep, yrad + nannuli * sep) + 3;
	ny = 2 * max(xrad + nannuli * sep, yrad + nannuli * sep) + 3;
	lx = x - (nx / 2);
	ly = y - (ny / 2);
	cdl_initMarkPos (mk, nx, ny, lx, ly);

	if (cdl_debug)
	    printf("[cdl_markElAnn] (%d,%d) R=(%d,%d) pa=%g N=%d S=%d col=%d\n",
		x, y, xrad, yrad, ang, nannuli, sep, color);

	/* Get the pixel region we'll be marking. */
	pix = (uchar *) malloc (nx * ny);
	if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	/* Save the image pixels in the display list for erasure later. */
	mk->refpix = (uchar *) malloc (nx * ny);
	mk->markpix = (uchar *) malloc (nx * ny);
	bcopy (pix, mk->refpix, nx * ny);
	cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

	/* Edit the pixels with the requested color and save the raster. */
	cdl_doEllipAnnuliMark (x, y, xrad, yrad, ang, nannuli,
	    sep, color, pix, lx, ly, nx, ny);
	mk->markpix = pix;

	/* Write the region back to the display. */
	if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
	    return (ERR);

	return (OK);
}



/*  CDL_MARKTEXT -- Draw a text string on the display at the given location.
 */

int
cdl_markText (cdl, x, y, str, size, angle, color)
CDLPtr  cdl;                            /* package ptr          */
int     x, y;                           /* marker position      */
char    *str;                           /* text string          */
float   size;                           /* size scale factor    */
float	angle;				/* text rotation angle	*/
int     color;                          /* marker color         */
{
        register int lx, ly, nx, ny;
        register int len, space = 1;
        uchar    *pix = NULL;
        MarkerPtr mk = cdl_initTextMarker (x, y, str, size, color);
        extern  int     cdl_getPixRegion();
        extern  void    cdl_insertMarker(), cdl_doLineInSubRas();

        if (cdl_debug)
            printf ("[cdl_markText] (%d,%d) str='%s' size=%g color=%d ang=%g\n",
                x, y, str, size, color, angle);

        len = strlen (str);
	if (angle == 0.0) {
	    /* Optimize a bit for horizontal text. */
            lx = x;
            ly = y - 5;
            nx = len * FONT_WIDTH * (size * FONT_SCALE) + len * space;
            ny = FONT_HEIGHT * (size * FONT_SCALE) + 10;
	} else {
	    /* Compute the four corners of the rotated and scaled text.  Then
	     * define the raster needed to get all of the text in the box.
	     */
	    int 	xc[4], yc[4];
	    int		 xmin, xmax, ymin, ymax;
	    double 	 coso, sino;
	    float 	 cw, ch;

            cw = FONT_WIDTH * (size * FONT_SCALE);
            ch = FONT_HEIGHT * (size * FONT_SCALE);
            coso = cos ((double) (-angle * M_PI / 180.));
            sino = sin ((double) (-angle * M_PI / 180.));

	    xc[0] = x;					    /* LL corner */
	    yc[0] = y;
	    xc[1] = x + ch * sino;			    /* UL corner */
	    yc[1] = y + ch * coso;
	    xc[2] = x + (len*cw+len) * coso + ch * sino;    /* UR corner */
	    yc[2] = y - (len*cw+len) * sino + ch * coso;
	    xc[3] = x + (len*cw+len) * coso;    	    /* LR corner */
	    yc[3] = y - (len*cw+len) * sino;
	    cdl_minmax (xc, 4, &xmin, &xmax);
	    cdl_minmax (yc, 4, &ymin, &ymax);

	    lx = xmin,  ly = ymin;
	    nx = xmax - xmin + 1;
	    ny = ymax - ymin + 1;
	}
        cdl_initMarkPos (mk, nx, ny, lx, ly);

        /* Get the pixel region we'll be marking. */
        pix = (uchar *) malloc (nx * ny);
        if (cdl_getPixRegion (cdl, lx, ly, nx, ny, pix))
            return (ERR);

        /* Save the image pixels in the display list for erasure later. */
        mk->refpix = (uchar *) malloc (nx * ny);
        mk->markpix = (uchar *) malloc (nx * ny);
        bcopy (pix, mk->refpix, nx * ny);
        cdl_insertMarker (cdl, DLTail[cdl->frame-1], mk);

        /* Edit the pixels with the requested color and save the raster. */
        cdl_doTextMarker (x, y, str, size, angle, color, pix, lx, ly, nx, ny);
        mk->markpix = pix;

        /* Write the region back to the display. */
        if (cdl_writeSubRaster (cdl, lx, ly, nx, ny, pix))
            return (ERR);

        return (OK);
}



/*  CDL_DELETEMARK -- Delete the overlay mark whose center is closest to
 *  the given position.
 */

int
cdl_deleteMark (cdl, x, y)
CDLPtr	cdl;				/* package ptr		*/
int	x, y;				/* marker position	*/
{
	MarkerPtr	mk, next, back;

	if (cdl_debug)
	    printf ("[cdl_deleteMark] (%d,%d)\n", x, y);

	if (DLHead[cdl->frame] == (MarkerPtr) NULL)
	    return (OK);

	/* Get the marker closest to the given point. */
	mk = cdl_findNearest (DLHead[cdl->frame-1], x, y);

        /* Write the region back to the display. */
        if (cdl_writeSubRaster(cdl, mk->lx, mk->ly, mk->nx, mk->ny, mk->refpix))
            return (ERR);

        /* For point markers we may need to also delete an associated label. */
        if (mk->type == MK_POINT && mk->number > 0 && mk->next != NULL) {
	    next = mk->next;
            if (cdl_writeSubRaster(cdl, next->lx, next->ly,
                next->nx, next->ny, next->refpix))
                    return (ERR);
            cdl_removeMarker (cdl, next);
        } else if (mk->type == MK_TEXT && mk->back != NULL) {
	    /*  If it's a text marker, see if the previous on was a point
	     *  with a label.
	     */
	    back = mk->back;
	    if (back->type == MK_POINT && back->number > 0) {
                if (cdl_writeSubRaster(cdl, back->lx, back->ly,
                    back->nx, back->ny, back->refpix))
                        return (ERR);
                cdl_removeMarker (cdl, back);
	    }
	}
 
	/* Delete the marker from the display list. */
	cdl_removeMarker (cdl, mk);

	return (OK);
}


/*  CDL_CLEAROVERLAY -- Erase all marks in the current display list.  Work
 *  from the tail of the list so we handle overlapping marks correctly.
 */

int
cdl_clearOverlay  (cdl)
CDLPtr	cdl;				/* package ptr		*/
{
	MarkerPtr	mk, back;

	if (cdl_debug)
	    printf ("[cdl_clearOverlay]\n");

	/* To clear the overlay we simply walk the display list and delete
	 * each point.
	 */
	for (mk = DLTail[cdl->frame-1]; mk != (MarkerPtr) NULL; mk = back) {
            /* Write the region back to the display. */
            if (cdl_writeSubRaster (cdl, mk->lx, mk->ly, mk->nx, mk->ny,
		mk->refpix))
                    return (ERR);

	    /* Delete the marker from the display list.  */
	    back = mk->back;
	    cdl_removeMarker (cdl, mk);
	}
	DLHead[cdl->frame-1] = DLTail[cdl->frame-1] = (MarkerPtr) NULL;

        return (OK);
}


/*  CDL_REDRAWOVERLAY -- Redraw all marks in the current display list.
 */

int
cdl_redrawOverlay  (cdl)
CDLPtr	cdl;				/* package ptr		*/
{
	MarkerPtr	mk, next;

	if (cdl_debug)
	    printf ("[cdl_redrawOverlay]\n");

	/*  Just walk the list redisplaying the markers.  */
	for (mk = DLHead[cdl->frame-1]; mk != (MarkerPtr) NULL; mk = mk->next)
            /* Write the region back to the display. */
            if (cdl_writeSubRaster (cdl, mk->lx, mk->ly, mk->nx, mk->ny,
		mk->markpix))
                    return (ERR);

	return (OK);
}



/* -------------------
 * PRIVATE PROCEDUURES
 * ------------------- */


/*  CDL_GETPIXREGION -- Given a corner position and size, return a subraster
 *  of the image pixels in that region. 
 */

static int
cdl_getPixRegion (cdl, lx, ly, nx, ny, ras)
CDLPtr	cdl;				/* package ptr		*/
int	lx, ly;				/* corner position	*/
int	nx, ny;				/* corner position	*/
uchar	*ras;				/*			*/
{
	register uchar	*ip;
	register int i, j, p, y, frame = cdl->frame;

	if (ras == NULL)
	    ras = (uchar *) malloc (nx * ny);

	/* Do a server query to get the pixels. */
	return (cdl_readSubRaster (cdl, lx, ly, nx, ny, &ras));
}


/*  CDL_DOPOINTMARK -- Edit a given raster with the requested mark.
 */

static void
cdl_doPointMark (color, type, pix, sz)
int	color;				/* overlay color	*/
int	type;				/* type of mark to draw */
uchar	*pix;				/* edit region raster	*/
int	sz;				/* mark size		*/
{
	register int i, j, k, npts, center, cpix, fill = 0;
	int 	x, y, width, xp[4], yp[4];

	/* Save the central pixel value, we don't overwrite that. */
	x = y = (sz / 2);
	center = ((sz / 2) * sz) + (sz / 2);
	cpix = pix[center];

	/* Check flags */
	if (type & M_POINT) {
	    cpix = color;
	    pix[center] = cpix;
	}
	if (type & M_FILL)
	    fill++;

	/* Draw the types requested. */
	if (type & M_BOX)
	     cdl_doBoxMark (color, fill, pix, sz, sz);

	if (type & M_PLUS || type & M_STAR) {
	    j = (y * sz);
	    for (i=0; i < sz; i++)
	        pix[j+i] = color;
	    j = ((sz - 1) * sz) + x;
	    for (i=0; i < sz; i++, j -= sz)
		pix[j] = color;
	}

	if (type & M_HLINE || type & M_HBLINE) {
	    npts = (sz - (sz / 3)) / 2;
	    if (type & M_HBLINE) {
	        j = ((y - 1) * sz);
		width = 3;
	    } else {
	        j = (y * sz);
		width = 1;
	    }
	    for (k=0; k < width; k++, j += sz) {
	        for (i=0; i < npts; i++) {
		    pix[j+i] = color;
		    pix[j+sz-i-1] = color;
	        }
	    }
	}

	if (type & M_VLINE || type & M_VBLINE) {
	    npts = (sz - (sz / 3)) / 2;
	    if (type & M_VBLINE) {
	        for (k=-1; k < 2; k++) {
	            j = ((sz - 1) * sz) + x;
	            for (i=0; i < npts; i++, j -= sz) {
		        pix[j+k] = color;
		        pix[(i*sz)+x+k] = color;
	            }
	        }
	    } else {
	        j = ((sz - 1) * sz) + x;
	        for (i=0; i < npts; i++, j -= sz) {
		    pix[j] = color;
		    pix[(i*sz)+x] = color;
	        }
	    }
	}

	if (type & M_CROSS || type & M_STAR) {
	    for (i=0, j=((sz-1)*sz); i < sz; i++, j-=sz) {
		pix[(i*sz)+i] = color; 			/* main diagonal */
		pix[j+i] = color;			/* off diagonal	 */
	    }
	}

	if (type & M_DIAMOND) {
            xp[0] = x - (sz/2);     yp[0] = y;
            xp[1] = x;              yp[1] = y + (sz/2);
            xp[2] = x + (sz/2);     yp[2] = y;
            xp[3] = x;              yp[3] = y - (sz/2);
            cdl_doPolygonMark (color, fill, xp, yp, 4, pix, 0, 0, sz, sz, True);
	    pix[x] = color;
	}

	if (type & M_CIRCLE)
	    cdl_doCircleMark(x, y, (sz/2)+1, color, fill, pix, 0, 0, 
		sz, sz);

	/* Restore the central pixel value, we don't usually overwrite that.
	 */
	if (!fill && !(type & M_POINT))
	    pix[center] = cpix;
}


/*  CDL_DOLINEMARK -- Edit a given raster with the requested mark.
 */

static void
cdl_doLineMark (color, pix, xs, ys, xe, ye)
int	color;				/* overlay color	*/
uchar	*pix;				/* edit region raster	*/
int	xs, ys, xe, ye;			/* line endpoints	*/
{
	register int i, j, y, nx, ny, np,lx, ly;

	nx = ABS (xe - xs) + 1;
	ny = ABS (ye - ys) + 1;
	lx = min(xs,xe);
	ly = min(ys,ye);
	np = (int) sqrt (nx * nx + ny * ny);

	if (xs == xe || ys == ye) {
	    /* Handle the case where a line falls on a line or column. */
	    for (i=0; i < np; i++)
		pix[i] = color;

	} else if ((xe - xs) == (ye - ye)) {
	    /* Handle the special case where a line is a diagonal. */
	    if (ye > ys)
	        for (i=0; i < nx; i++)
	            pix[(i*nx)+i] = color; 		/* main diagonal */
	    else
	        for (i=0, j=((ny-1)*nx); i < nx; i++, j-=nx)
	            pix[j+i] = color;			/* off diagonal	 */

	} else {
	    /*  Handle a line between arbitrary endpoints.  */
	    cdl_doLineInSubRas (pix, color, xs, xe, ys, ye,
		lx, ly, nx ,ny);
	}
}


/*  CDL_DOBOXMARK -- Edit a given raster with the requested mark.
 */

static void
cdl_doBoxMark (color, fill, pix, nx, ny)
int	color;				/* overlay color	*/
int	fill;				/* fill the box?	*/
uchar	*pix;				/* edit region raster	*/
int	nx, ny;				/* box size		*/
{
	register int i, j, np = (nx * ny);

	/* Edit the pixels with the requested color. */
	if (fill) {
	    /* Fill the entire rectangle with the color. */
	    for (i=0; i < np ; i++)
		pix[i] = color;
	} else {
	    /* Change just the borders of the box to requested color. */
	    for (i=0, j = (ny - 1) * nx; i < nx; i++, j++)
		pix[i] = pix[j] = color;
	    for (i=0, j = (nx - 1); i < np; i+=nx, j+=nx)
	        pix[i] = pix[j] = color;
	}
}


/*  CDL_DOPOLYGONMARK -- Edit a given raster with the requested mark.
 */

static void
cdl_doPolygonMark (color, fill, x, y, npts, pix, lx, ly, nx, ny, close_poly)
int	color;				/* overlay color	*/
int	fill;				/* fill polygon 	*/
int	*x, *y;				/* coordinate arrays	*/
int	npts;				/* npts to plot		*/
uchar	*pix;				/* edit region raster	*/
int	lx, ly;				/* subraster corner	*/
int	nx, ny;				/* subraster size  	*/
int	close_poly;			/* close the polygon 	*/
{
	register int i, nseg = (npts - 1);

	if (cdl_debug)
	    printf ("[cdl_doPolygonMark] npts=%d lx=%d ly=%d nseg=%d\n",
	 	npts, lx, ly, nseg);

	/* Connect all the vertices. */
	for (i=0; i < nseg; i++)
	    cdl_doLineInSubRas (pix, color, x[i], x[i+1],
		y[i], y[i+1], lx, ly, nx, ny);

	/* Now connect the last point to the first to close the polygon.  */
	if (close_poly) {
	    if (x[0] != x[npts] && y[0] != y[npts])
	        cdl_doLineInSubRas (pix, color, x[0], x[nseg],
		    y[0], y[nseg], lx, ly, nx, ny);

	    /* Flood fill the polygon if requested. */
	    if (fill)
	        cdl_fillArea (pix, nx, ny, color);
	}
}


/*  CDL_DOCIRCLEMARK -- Edit a given raster with the requested mark.
 */

static void
cdl_doCircleMark (x, y, radius, color, fill, pix, lx, ly, nx, ny)
int	x, y;				/* center point		*/
int 	radius;				/* circle radius	*/
int	color;				/* overlay color	*/
int	fill;				/* fill the box?	*/
uchar	*pix;				/* edit region raster	*/
int	lx, ly;				/* subraster corner	*/
int	nx, ny;				/* box size		*/
{
  	register int i, np;
  	int xp[N_CIRCLE_PTS+1], yp[N_CIRCLE_PTS+1];

	/* Get the coordinates for drawing the circle. */
	cdl_getCircleCoords (x, y, radius, xp, yp);

	/* Connect the dots to draw the circle. */
	cdl_doPolygonMark (color, fill, xp, yp, N_CIRCLE_PTS,
	    pix, lx, ly, nx, ny, True);
}


/*  CDL_DOCIRCANNULIMARK -- Edit a given raster with the requested mark. 
 */

static void
cdl_doCircAnnuliMark(x, y, radius, nannuli, sep, color, pix, lx, ly, nx, ny)
int	x, y;				/* center point		*/
int 	radius;				/* circle radius	*/
int	nannuli;			/* No. of annuli	*/
int	sep;				/* annulus separation	*/
int	color;				/* overlay color	*/
uchar	*pix;				/* edit region raster	*/
int	lx, ly;				/* subraster corner	*/
int	nx, ny;				/* box size		*/
{
  	register int i, j, nseg;
  	int xp[N_CIRCLE_PTS+1], yp[N_CIRCLE_PTS+1];

	/* For each annulus radius draw the circle. */
	for (i=0; i <= nannuli; i++)
	    cdl_doCircleMark (x, y, radius + (i* sep), color, False,
		pix, lx, ly, nx, ny);
}


/*  CDL_DOELLIPSEMARK -- Edit a given raster with the requested mark.
 */

static void
cdl_doEllipseMark (color, fill, xc, yc, xrad, yrad, ang, pix, lx, ly, nx, ny)
int	color;				/* overlay color	*/
int	fill;				/* fill the box?	*/
int	xc, yc;				/* ellipse center	*/
int	xrad, yrad;			/* semimajor axes	*/
float	ang;				/* rotation angle	*/
uchar	*pix;				/* edit region raster	*/
int	lx, ly;				/* subraster corner	*/
int	nx, ny;				/* box size		*/
{
  	register int i, nseg;
  	int xp[N_ELLIPSE_PTS+1], yp[N_ELLIPSE_PTS+1];

	/* Get the coordinates for drawing the ellipse. */
	cdl_getEllipseCoords (xc, yc, xrad, yrad, ang, xp, yp);

	/* Connect the dots to draw the ellipse.  */
	cdl_doPolygonMark (color, fill, xp, yp, N_ELLIPSE_PTS,
	    pix, lx, ly, nx, ny, True);
}


/*  CDL_DOELLIPANNULIMARK -- Edit a given raster with the requested mark. 
 */

static void
cdl_doEllipAnnuliMark(x, y, xrad, yrad, ang, nannuli, sep, color,
	pix, lx, ly, nx, ny)
int	x, y;				/* ellipse center	*/
int	xrad, yrad;			/* semimajor axes	*/
float	ang;				/* rotation angle	*/
int	nannuli;			/* No. of annuli	*/
int	sep;				/* annulus separation	*/
int	color;				/* overlay color	*/
uchar	*pix;				/* edit region raster	*/
int	lx, ly;				/* subraster corner	*/
int	nx, ny;				/* box size		*/
{
  	register int i, j, nseg, xr, yr;
  	int xp[N_ELLIPSE_PTS+1], yp[N_ELLIPSE_PTS+1];

	/* For each annulus compute the ellipse and draw it... */
	for (i=0; i <= nannuli; i++) {
	    xr = xrad + (i * sep);
	    yr = yrad + (i * sep);
	    cdl_doEllipseMark (color, False, x, y, xr, yr, ang,
		pix, lx, ly, nx, ny);
	}
}


/*  CDL_DOTEXTMARKER -- Draw a text string to the display.
 */

int
cdl_doTextMarker (x, y, string, size, angle, color, pix, lx, ly, nx, ny)
int     x, y;				/* LL text position	*/
char    *string;			/* string to draw	*/
float   size;				/* relative scale size	*/
float	angle;				/* text rotation angle	*/
int     color;				/* overlay color	*/
uchar   *pix;                           /* edit region raster   */
int     lx, ly;                         /* subraster corner     */
int     nx, ny;                         /* box size             */
{
        char    *sp = string;
        int     mx, my, stroke, tab1, tab2, i, j, pen, len;
        int     x1, x2, y1, y2;
        register int cw, ch, cx, cy;
        float   px, py;
	double	coso, sino;
        int     bitupk();

	coso = cos ((double) (-angle * M_PI / 180.));
	sino = sin ((double) (-angle * M_PI / 180.));

        /* Loop over each character in the string. */
        len = strlen (string);
        x1 = cx = x;
        y1 = cy = y;
        cw = FONT_WIDTH * (size * FONT_SCALE);
        ch = FONT_HEIGHT * (size * FONT_SCALE);
        for (j = 0; j < len; j++, sp++) {

            if (*sp < CHARACTER_START || *sp > CHARACTER_END)
                i = '?' - CHARACTER_START;
            else
                i = *sp - CHARACTER_START;

            tab1 = chridx[i] - 1;
            tab2 = chridx[i+1] - 1;

            for (i=tab1; i <= tab2; i++) {
                stroke = chrtab[i];
                px  = (float) bitupk (stroke, COORD_X_START, COORD_X_LEN);
                py  = (float) bitupk (stroke, COORD_Y_START, COORD_Y_LEN);
                pen = bitupk (stroke, COORD_PEN_START, COORD_PEN_LEN);

                /* Scale size of character.  */
                px = px * (size * FONT_SCALE);
                py = py * (size * FONT_SCALE);

                /* Shift and rotate.
		 */
                mx = cx + px * coso + py * sino;
                my = cy - px * sino + py * coso;

                if (pen == 0) {
                    /* Move start point. */
                    x1 = mx;
                    y1 = my;
                } else {
                    /* Draw to this point and update. */
                    x2 = mx;
                    y2 = my;
                    cdl_doLineInSubRas (pix, color, x1, x2, y1, y2,
                        lx, ly, nx, ny);
                    x1 = x2;
                    y1 = y2;
                }
            }

	    /* Update the character position. */
	    cx += cw * coso;
	    cy -= ch * sino;
        }
}




/*  ------------------------
 *  Display List Procedures
 *  ------------------------*/



/*  CDL_FREEDISPLAYLIST -- Free up the named display list.
 */

int
cdl_freeDisplayList  (cdl, head)
CDLPtr	cdl;				/* package ptr		*/
{
	MarkerPtr	mk, next;

	if (cdl_debug)
	    printf ("[cdl_freeDisplayList]\n");

	/*  Just walk the list freeing the markers.  */
	for (mk = DLHead[cdl->frame-1]; mk != (MarkerPtr) NULL; mk = mk->next)
            cdl_removeMarker (cdl, mk);

	return (OK);
}


/*  CDL_INSERTMARKER -- Insert a marker in the display list after the given
 *  entry.  The display list is maintained as a doubly linked list of markers.
 */

static void
cdl_insertMarker (cdl, back, new)
CDLPtr		cdl;			/* package ptr		*/
MarkerPtr	back;			/* insert point		*/
MarkerPtr	new;			/* marker to insert	*/
{
	MarkerPtr	tmp;

	/* If the back pointer is null it means we're creating the head. */
	if (back == (MarkerPtr) NULL) {
	    DLHead[cdl->frame-1] = DLTail[cdl->frame-1] = new;
	    new->back = new->next = (MarkerPtr) NULL;

	} else {
	    /* Otherwise reset the pointers to insert the marker. */
	    tmp = back->next;
	    back->next = new;
	    new->back = back;

	    /* If the beck marker pointed at something, take care of that. */
	    if (tmp) {
	        new->next = tmp;
	        tmp->back = new;
	    } else {
	        new->next = (MarkerPtr) NULL;
		DLTail[cdl->frame-1] = new;
	    }
	}
}


/*  CDL_REMOVEMARKER -- Remove a marker from the display list.
 */

static void
cdl_removeMarker (cdl, mk)
CDLPtr		cdl;			/* package ptr		*/
MarkerPtr	mk;			/* marker to delete	*/
{
	if (mk->back) 			/* handle the back ptr  */
	    mk->back->next = mk->next;
	else
	    DLHead[cdl->frame-1] = mk->next;
	if (mk->next) 			/* handle the next ptr  */
	    mk->next->back = mk->back;
	else
	    DLTail[cdl->frame-1] = mk->back;

	/* Now free up the marker and internal pointers. */
	if (mk->str)     free ((uchar *) mk->str);
	if (mk->refpix)  free ((uchar *) mk->refpix);
	if (mk->markpix) free ((uchar *) mk->markpix);
	if (mk->xp) 	 free ((uchar *) mk->xp);
	if (mk->yp) 	 free ((uchar *) mk->yp);
	free ((MarkerPtr) mk);
}


/*  CDL_FINDNEAREST -- Find the marker in the display list that is closest
 *  to the indicated point.  For markers with a defined center we check the
 *  distance to the central point.  For polylines and polygons we check for
 *  a vertex ir knot point that is closest.  The marker is then returned.
 */

static MarkerPtr
cdl_findNearest (head, x, y)
MarkerPtr	head;			/* display list head	*/
int		x, y;			/* pixel coords		*/
{
	MarkerPtr	mk, nearest;
	register float  dmin = 10.0e9, mk_dist, dist = 10.0e9, A, C;
	register int	i;

	if (head == (MarkerPtr) NULL)
	    return ((MarkerPtr) NULL);

	nearest = head;
	for (mk = head; mk ; mk = mk->next) {
	    switch (mk->type) {
	    case MK_POLYLINE:
	    case MK_POLYGON:
		/* Find the nearest point in the poly. */
		for (i=0; i < mk->npts; i++) {
		    mk_dist = sqrt ((double)(ABS(x - mk->x) * ABS(x - mk->x) +
			        	     ABS(y - mk->y) * ABS(y - mk->y)) );
	  	    if (mk_dist < dmin)
		        dmin = mk_dist;
		}
		mk_dist = dmin;
		break;

	    case MK_BOX:
		/* See if it's between the verticals... */
		if (x >= mk->xp[0] && x <= mk->xp[1])
		    mk_dist = min (ABS(y - mk->yp[0]), ABS(y - mk->yp[1]));
		else if (y >= mk->yp[0] && y <= mk->yp[1])
		    /* See if it's between the horizontals... */
		    mk_dist = min (ABS(x - mk->xp[0]), ABS(x - mk->xp[1]));
		break;

	    case MK_LINE:
		A = (float) ((mk->yp[1] - mk->yp[0]) / (mk->xp[1] - mk->xp[0]));
		C = mk->yp[0] - A * x;
		mk_dist = ABS( (A * x - y + C) / sqrt (A * A + 1));
		break;

	    case MK_POINT:
	    case MK_CIRCLE:
	    case MK_CIRCANN:
	    case MK_ELLIPSE:
	    case MK_ELLIPANN:
	    case MK_TEXT:
		/* Get the distance from the point to the center. */
		mk_dist = sqrt ((double)(ABS(x - mk->x) * ABS(x - mk->x) +
			                 ABS(y - mk->y) * ABS(y - mk->y)) );
	    }
	    if (mk_dist < dist) {
		dist = mk_dist;
	        nearest = mk;
	    }

	    if (cdl_debug) 
		printf ("Marker type=%2d  dist=%g  mk_dist=%g  N=%d\n",
		    mk->type, dist, mk_dist,nearest->type);
	}
	return (nearest);
}


/*  CDL_PRINTDISPLAYLIST -- Print the display list to a file.
 */

static void
cdl_printDisplayList (fd, head)
FILE		*fd;			/* file descriptor	*/
MarkerPtr	head;			/* display list head	*/
{
	MarkerPtr	mk;

	for (mk = head; mk ; mk = mk->next) {
	    switch (mk->type) {
	    case MK_POINT:
		printf ("point\t\t");
		break;
	    case MK_LINE:
		printf ("line\t\t");
		break;
	    case MK_BOX:
		printf ("box\t\t");
		break;
	    case MK_POLYLINE:
		printf ("polyline\t\t");
		break;
	    case MK_POLYGON:
		printf ("polygon\t\t");
		break;
	    case MK_CIRCLE:
		printf ("circle\t\t");
		break;
	    case MK_CIRCANN:
		printf ("circann\t\t");
		break;
	    case MK_ELLIPSE:
		printf ("ellipse\t\t");
		break;
	    case MK_ELLIPANN:
		printf ("ellipann\t\t");
		break;
	    case MK_TEXT:
		printf ("text\t\t");
		break;
	    }
	    printf ("mk=%d back=%10d  next=%10d\n", mk, mk->back, mk->next);
	}	
}


/*  CDL_INITPOINTMARKER -- Allocate and initialize the space for a point marker.
 */

static MarkerPtr
cdl_initPointMarker (x, y, number, size, type, color)
int     x, y;                           /* marker position      */
int     number;                         /* if > 0, label value  */
int     size;                           /* marker size (pixels) */
int     type;                           /* type to draw         */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type    	= MK_POINT;
	mk->x       	= x;
	mk->y       	= y;
	mk->number  	= number;
	mk->size    	= size;
	mk->pt_type 	= type;
	mk->color   	= color;
	return mk;
}


/*  CDL_INITLINEMARKER -- Allocate and initialize the space for a line marker.
 */

static MarkerPtr
cdl_initLineMarker (x1, y1, x2, y2, color)
int     x1, y1;                         /* line start points    */
int     x2, y2;                         /* line end points      */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type	= MK_LINE;
	mk->x 		= x1;
	mk->y 		= y1;
	mk->xp 		= (int *) malloc (2 * sizeof(int));
	mk->xp[0] 	= x1;
	mk->xp[1] 	= x2;
	mk->yp 		= (int *) malloc (2 * sizeof(int));
	mk->yp[0] 	= y1;
	mk->yp[1] 	= y2;
	mk->color	= color;
	return mk;
}


/*  CDL_INITBOXMARKER -- Allocate and initialize the space for a box marker.
 */

static MarkerPtr
cdl_initBoxMarker (lx, ly, ux, uy, fill, color)
int     lx, ly;                         /* LL corner points     */
int     ux, uy;                         /* UR corner points     */
int     fill;                           /* fill rectangle?      */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type	= MK_BOX;
	mk->x 		= lx;
	mk->y 		= ly;
	mk->xp 		= (int *) malloc (2 * sizeof(int));
	mk->xp[0] 	= lx;
	mk->xp[1] 	= ux;
	mk->yp 		= (int *) malloc (2 * sizeof(int));
	mk->yp[0] 	= ly;
	mk->yp[1] 	= uy;
	mk->fill	= fill;
	mk->color	= color;
	return mk;
}


/*  CDL_INITPOLYLINEMARKER -- Allocate and initialize the space for a 
 *  polyine  marker.
 */

static MarkerPtr
cdl_initPolylineMarker (x, y, npts, color)
int     *x, *y;               		/* vertex points        */
int     npts;                           /* number of points     */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type	= MK_POLYLINE;
	mk->xp 		= (int *) malloc (npts * sizeof(int));
	mk->yp 		= (int *) malloc (npts * sizeof(int));
	bcopy (x, mk->xp, npts * sizeof(int));
	bcopy (y, mk->yp, npts * sizeof(int));
	mk->x 		= mk->xp[0];
	mk->y 		= mk->yp[0];
	mk->npts	= npts;
	mk->color	= color;
	return mk;
}


/*  CDL_INITPOLYGONMARKER -- Allocate and initialize the space for a 
 *  polygon marker.
 */

static MarkerPtr
cdl_initPolygonMarker (x, y, npts, fill, color)
int     x, y;             		/* vertex points        */
int     npts;                           /* number of corners    */
int     fill;                           /* fill polygon?        */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type	= MK_POLYGON;
	mk->xp 		= (int *) malloc (npts * sizeof(int));
	mk->yp 		= (int *) malloc (npts * sizeof(int));
	bcopy (x, mk->xp, npts * sizeof(int));
	bcopy (y, mk->yp, npts * sizeof(int));
	mk->x 		= mk->xp[0];
	mk->y 		= mk->yp[0];
	mk->npts	= npts;
	mk->fill	= fill;
	mk->color	= color;
	return mk;
}


/*  CDL_INITCIRCLEMARKER -- Allocate and initialize the space for a
 *  circle marker.
 */

static MarkerPtr
cdl_initCircleMarker (x, y, radius, fill, color)
int     x, y;                           /* center position      */
int     radius;                         /* radius of circle     */
int     fill;                           /* fill circle?         */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type	= MK_CIRCLE;
	mk->x       	= x;
	mk->y       	= y;
	mk->radius	= radius;
	mk->fill	= fill;
	mk->color	= color;
	return mk;
}


/*  CDL_INITCIRCANNMARKER -- Allocate and initialize the space for a 
 *  circular annuli marker.
 */

static MarkerPtr
cdl_initCircAnnMarker (x, y, radius, nannuli, sep, color)
int     x, y;                           /* center position      */
int     radius;                         /* radius of 1st annulus*/
int     nannuli;                        /* no. of annuli        */
int     sep;                            /* annuli sep (pixels)  */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type	= MK_CIRCANN;
	mk->x       	= x;
	mk->y       	= y;
	mk->radius	= radius;
	mk->nannuli	= nannuli;
	mk->sep		= sep;
	mk->color	= color;
	return mk;
}


/*  CDL_INITELLIPSEMARKER -- Allocate and initialize the space for a 
 *  ellipse marker.
 */

static MarkerPtr
cdl_initEllipseMarker (x, y, xrad, yrad, ang, fill, color)
int     x, y;                           /* center position      */
int     xrad, yrad;                     /* x and y radii        */
float   ang;                            /* position angle (deg) */
int     fill;                           /* fill ellipse?        */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type	= MK_ELLIPSE;
	mk->x       	= x;
	mk->y       	= y;
	mk->xrad	= xrad;
	mk->yrad	= yrad;
	mk->ang 	= ang;
	mk->fill	= fill;
	mk->color 	= color;
	return mk;
}


/*  CDL_INITELLANNMARKER -- Allocate and initialize the space for a 
 *  elliptical annuli marker.
 */

static MarkerPtr
cdl_initEllAnnMarker (x, y, xrad, yrad, ang, nannuli, sep, color)
int     x, y;                           /* center position      */
int     xrad, yrad;                     /* radius of 1st annulus*/
float   ang;                            /* rotation angle       */
int     nannuli;                        /* no. of annuli        */
int     sep;                            /* annuli sep (pixels)  */
int     color;                          /* marker color         */
{
	MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

	mk->type	= MK_ELLIPANN;
	mk->x       	= x;
	mk->y       	= y;
	mk->xrad	= xrad;
	mk->yrad	= yrad;
	mk->ang 	= ang;
	mk->nannuli	= nannuli;
	mk->sep		= sep;
	mk->color 	= color;
	return mk;
}


/*  CDL_INITTEXTMARKER -- Allocate and initialize the space for a text marker.
 */

static MarkerPtr
cdl_initTextMarker (x, y, string, size, color)
int     x, y;                           /* marker position      */
char    *string;                        /* string to draw       */
float   size;                           /* marker size (pixels) */
int     color;                          /* marker color         */
{
        MarkerPtr mk = (MarkerPtr) calloc (1, sizeof(struct Marker));

        mk->type        = MK_TEXT;
        mk->x           = x;
        mk->y           = y;
        mk->txsize      = size;
        mk->color       = color;
        mk->str         = (char *) malloc (strlen(string) + 1);
        strcpy (mk->str, string);
        return mk;
}



static void
cdl_initMarkPos (mk, nx, ny, lx, ly)
MarkerPtr	mk;
int		nx, ny;
int		lx, ly;
{
	mk->nx = nx;
	mk->ny = ny;
	mk->lx = lx;
	mk->ly = ly;
}



/*  ------------------------
 *  Misc. Utility Procedures
 *  ------------------------*/

/*  CDL_DOLINEINSUBRAS --  Draw a line between two arbitrary endpoints in a 
 *  subraster.  Endpoints are given in image coordinates along with the
 *  subraster boundary, we translate to subraster coords before marking the
 *  pixel.  This allows us to get a subraster for a larger object such as
 *  polygon and use the line primitive to draw each vertex before writing
 *  back a single subraster rather than use the less efficient sequence of
 *  server I/O calls for each line segment.  Uses Bresenham's algorithm to
 *  get a smoother line.
 */

static void
cdl_doLineInSubRas (pix, color, x1, x2, y1, y2, lx, ly, nx, ny)
uchar	*pix;				/* edit region raster	*/
int	color;				/* overlay color	*/
int	x1, y1, x2, y2;			/* line endpoints	*/
int	lx, ly;				/* subraster LL corner  */
int	nx, ny;				/* subraster size  	*/
{
	register int  i, d, x, y;
    	register int  ax, ay, sx, sy, dx, dy;

	/* Use a Bresenham algorith do draw the line. */
    	dx = x2 - x1;    ax = ABS(dx) << 1;    sx = SGN(dx);
    	dy = y2 - y1;    ay = ABS(dy) << 1;    sy = SGN(dy);
	x = x1;
	y = y1;
    	if (ax > ay) {			/* X span is dominant */
            for (d = ay - (ax >> 1); x != x2; x += sx, d += ay) {
	        i = ((y-ly) * nx) + min(nx-1,(x-lx));
	        pix[i] = color;
            	if (d >= 0) {
                    y += sy;
                    d -= ax;
            	}
            }
    	} else {     			/* Y span is dominant */
            for (d = ax - (ay >> 1); y != y2; y += sy, d += ax) {
	        i = ((y-ly) * nx) + min(nx-1,(x-lx));
	        pix[i] = color;
            	if (d >= 0) {
                    x += sx;
                    d -= ay;
            	}
            }
    	}
}


/*  CDL_FILLAREA -- Scan the pixel array and fill the area enclosed by the
 *  given color, or replace with those pixels from the reference raster.
 *  We assume the area to be filled is a simple closed convex polygon.
 */

static void
cdl_fillArea (pix, nx, ny, color)
uchar	*pix;				/* edit region raster	*/
int	nx, ny;				/* subraster size  	*/
int	color;				/* overlay color	*/
{
	register int i, j, l, r;

	for (i=0; i < ny; i++) {
	    l = r  = 0;

	    /* Find the left endpoint. */
	    while (l < nx && pix[(i * nx) + l] != color)
		l++;
	    if (l != nx) {

	        /* Find the right endpoint. */
	        r = nx - 1;
	        while (r > l && pix[(i * nx) + r] != color)
		    r--;

	        if (r > l) {
	            /* Fill the area in between. */
	            for (j=l; j <= r; j++)
	  	        pix[(i*nx)+j] = color;
		}
	    }
	}
}


/*  CDL_GETCIRCLECOORDS --  Compute the coords for points needed to draw a
 *  circle of a given radius at the given center.
 */

static void
cdl_getCircleCoords (xcen, ycen, radius, x, y)
int	xcen, ycen;			/* circle center	*/
int	radius;				/* circle radius	*/
int	*x, *y;				/* output coords	*/
{
  	register int i, count, xoff, yoff, nseg;
  	register int a1, a2, a3, b1, b2, b3, b4;
        static double UnitCircleX[12] = {
            -0.06540312923014, -0.19509032201613,
            -0.32143946530316, -0.44228869021900,
            -0.55557023301960, -0.65934581510007,
            -0.75183980747898, -0.83146961230255,
            -0.89687274153269, -0.94693012949511,
            -0.98078528040323, -0.99785892323860
        };
        static double UnitCircleY[12] = {
            -0.99785892323860, -0.98078528040323,
            -0.94693012949511, -0.89687274153269,
            -0.83146961230255, -0.75183980747898,
            -0.65934581510007, -0.55557023301960,
            -0.44228869021900, -0.32143946530316,
            -0.19509032201613, -0.06540312923014
        };  

 
  	/*  Set up array indexes for 8 fold symetry  */
  	/*  8 indexes radiating both ways from each of the four axes  */
  	a1 = N_CIRCLE_PTS / 4;         /*  (1 * cnt) / 4  */
  	a2 = a1 + a1;                  /*  (2 * cnt) / 4  */
  	a3 = a2 + a1;                  /*  (3 * cnt) / 4  */
  	b1 = a1 - 1;
  	b2 = a2 - 1;
  	b3 = a3 - 1;
  	b4 = a3 + a1 - 1;              /*  (4 * cnt) / 4 - 1  */

  	/*  Calculate points on circumference for 1/8th of circle  */
  	/*  Apply to each of 8 pairs  */
  	count = N_CIRCLE_PTS / 8;
  	for( i=0; i <= count; i++ ) {
    	    xoff    = UnitCircleX[i] * radius + 0.5;
    	    yoff    = UnitCircleY[i] * radius + 0.5;
    	    x[i]    = xcen + xoff;
    	    y[i]    = ycen + yoff;
    	    x[a1+i] = xcen + yoff;
    	    y[a1+i] = ycen - xoff;
    	    x[a2+i] = xcen - xoff;
    	    y[a2+i] = ycen - yoff;
    	    x[a3+i] = xcen - yoff;
    	    y[a3+i] = ycen + xoff;
    	    x[b1-i] = xcen + yoff;
    	    y[b1-i] = ycen + xoff;
    	    x[b2-i] = xcen + xoff;
    	    y[b2-i] = ycen - yoff;
    	    x[b3-i] = xcen - yoff;
    	    y[b3-i] = ycen - xoff;
    	    x[b4-i] = xcen - xoff;
    	    y[b4-i] = ycen + yoff;
  	}
  	/*  Close the circle (end point same as starting point)  */
  	x[N_CIRCLE_PTS-1] = x[0];
  	y[N_CIRCLE_PTS-1] = y[0];  
}


/*  CDL_GETELLIPSECOORDS --  Compute the coords for points needed to draw a
 *  ellipse w/ given axes and angle at the desired center.
 */

static void
cdl_getEllipseCoords (xcen, ycen, xradius, yradius, rotang, x, y)
int	xcen, ycen;			/* ellipse center	*/
int	xradius, yradius;		/* ellipse axes		*/
float	rotang;				/* angle from X axis	*/
int	*x, *y;				/* output coords	*/
{
  	int 	loop;
  	float 	xoff, yoff;
	double	rotsin, rotcos;
  	register int i;
  	register int x1, x2, y1, y2;
  	register int a2, b1, b3;
	static double UnitEllipseX[16] = {
  	    -0.04906767432742, -0.14673047445536, 
	    -0.24298017990326, -0.33688985339222,
  	    -0.42755509343028, -0.51410274419322, 
	    -0.59569930449243, -0.67155895484702,
  	    -0.74095112535496, -0.80320753148064, 
	    -0.85772861000027, -0.90398929312344,
  	    -0.94154406518302, -0.97003125319454, 
	    -0.98917650996478, -0.99879545620517
	};
	static double UnitEllipseY[16] = {
  	    -0.99879545620517, -0.98917650996478, 
	    -0.97003125319454, -0.94154406518302,
  	    -0.90398929312344, -0.85772861000027, 
	    -0.80320753148064, -0.74095112535496,
  	    -0.67155895484702, -0.59569930449243, 
	    -0.51410274419322, -0.42755509343028,
  	    -0.33688985339222, -0.24298017990326, 
	    -0.14673047445536, -0.04906767432742
	}; 

  	/*  Set up array indexes for 8 fold symetry  */
  	/*  4 indexes radiating both ways from each of the two axis rays  */
  	a2 = N_ELLIPSE_PTS / 2;
  	b1 = a2 - 1;
  	b3 = N_ELLIPSE_PTS - 1;

	rotsin	= sin ((double) ((double)rotang * 0.017453293));
	rotcos	= cos ((double) ((double)rotang * 0.017453293));

  	/*  Recalculate points on circumference  */
  	/*  Apply to each of 4 pairs  */
  	loop = N_ELLIPSE_PTS / 4;
  	for( i=0; i < loop; i++ ) {
    	    /*  Calculate ray lengths for orthogonal case  */
    	    xoff    = UnitEllipseX[i] * xradius;
    	    yoff    = UnitEllipseY[i] * yradius;

    	    /*  Calculate offset of two points on the right (rotate)  */
    	    x2      = (xoff * rotcos) + (yoff * rotsin) + 0.5;
    	    y2      = (xoff * rotsin) - (yoff * rotcos) - 0.5;
    	    x1      = (xoff * rotcos) - (yoff * rotsin) - 0.5;
    	    y1      = (xoff * rotsin) + (yoff * rotcos) + 0.5;
    	    x[i]    = xcen + x1;
    	    y[i]    = ycen + y1;
    	    x[a2+i] = xcen - x1;
    	    y[a2+i] = ycen - y1;
    	    x[b1-i] = xcen + x2;
    	    y[b1-i] = ycen + y2;
    	    x[b3-i] = xcen - x2;
    	    y[b3-i] = ycen - y2;
  	}
  	/*  Close the circle (end point same as starting point)  */
  	x[N_ELLIPSE_PTS-1] = x[0];
  	y[N_ELLIPSE_PTS-1] = y[0]; 
}


/*  CDL_MINMAX -- Find the array extrema.  */

static void
cdl_minmax (array, npts, amin, amax)
int	*array;				/* array to check	*/
int	npts;				/* no of points		*/
int	*amin, *amax;			/* extremes		*/
{
	register int i;

	*amin = *amax = array[0];
	for (i=1; i < npts; i++) {
	    if (*amin > array[i])   *amin = array[i];
	    if (*amax < array[i])   *amax = array[i];
	}
}


/* BITUPK -- Unpack an unsigned integer bit field from a longword.
 */

unsigned bitmask[] = {  0,                      /* MACHDEP              */
        01,             03,             07,
        017,            037,            077,
        0177,           0377,           0777,
        01777,          03777,          07777,
        017777,         037777,         077777,
        0177777,        0377777,        0777777,
        01777777,       03777777,       07777777,
        017777777,      037777777,      077777777,
        0177777777,     0377777777,     0777777777,
        01777777777,    03777777777,    07777777777,
        017777777777,   037777777777
};

static int
bitupk (wordp, offset, nbits)
unsigned int   wordp;         /* longword to be examined              */
int    offset;                /* one-indexed offset of first bit      */
int    nbits;                 /* number of bits to be set             */
{
        return ((wordp >> (offset-1)) & bitmask[nbits]);
}

